package engine;

import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;

import game.Const;

/**
 * Couche d'affichage dynamique du jeu : construit l'image du jeu (carte,
 * joueurs, tirs, animation, ...) Plusieurs layers sont disponibles, recuperable
 * a partire du render
 * 
 * @author Mikael GUILLEMOT
 * @version 1.0
 */

public class Layer implements IRenderable {

	private List<IRenderable> renderableObjects = new ArrayList<>();
	private BufferedImage staticImage = null;

	/**
	 * ajoute un objet a dessiner lors du rendu du layer
	 * 
	 * @param object
	 */
	public void addRenderable(IRenderable object) {
		if (renderableObjects.contains(object))
			return;
		renderableObjects.add(object);
		updateStaticBuffer();
	}

	/**
	 * retire un objet des objets a dessiner lors du rendu du layer (si il existe)
	 * 
	 * @param object
	 */
	public void removeRenderable(IRenderable object) {
		renderableObjects.remove(object);
		updateStaticBuffer();
	}

	public void clear() {
		renderableObjects.clear();
		updateStaticBuffer();
	}

	@Override
	public void render(Graphics2D buffer) {
		if (staticImage == null)
			for (IRenderable r : renderableObjects)
				r.render(buffer);
		else
			buffer.drawImage(staticImage, 0, 0, null);
	}

	/**
	 * permet de definir le rendu de ce layer comme statique (l'image affichée n'est
	 * recalculée qu'en cas de modification des listes de rendu)
	 * 
	 * @param value
	 */
	public void setStaticRender(boolean isStatic) {
		if (isStatic) {
			staticImage = new BufferedImage(Const.W, Const.H, BufferedImage.TYPE_INT_ARGB);
			updateStaticBuffer();
		} else {
			staticImage = null;
		}

	}

	public void updateStaticBuffer() {
		if (staticImage != null) {
			Graphics2D staticBuffer = (Graphics2D) staticImage.getGraphics();
			for (IRenderable r : renderableObjects)
				r.render(staticBuffer);
		}
	}

	public int size() {
		return renderableObjects.size();
	}
}
