package game;

import java.util.ArrayList;
import java.util.List;

import engine.Commands;
import engine.IRenderable;
import engine.Render;
import game.objects.Mur;
import game.renderables.Briefing;
import game.renderables.LooseScreen;
import game.renderables.Selection;
import game.renderables.UI;
import game.renderables.WinScreen;
import game.objects.ITower;
import game.objects.Somb;
import game.objects.TowerBarbed;
import game.objects.TowerHummer;
import game.objects.TowerWatch;
import game.Const;

public class GameLoop extends Thread {

	private boolean alive = true;
	private Commands commands;
	private Render render;
	private long timer;
	private int status = Const.INIT;

	private long frame = 0;

	// valeurs de jeu
	private int waveDelai = 0;
	private int invasion = 0;
	private int cash = 0;

	// layers objets affichable tout au long du jeu
	private final IRenderable intro = new game.renderables.Intro();
	private final IRenderable briefing = new Briefing();
	// private IRenderable light = new LightingEffect();

	private final IRenderable win = new WinScreen();
	private final IRenderable loose = new LooseScreen();;

	private UI ui = new UI();
	private Selection select = new Selection();
	private Mur mur = new Mur();

	public GameLoop() {
		render = Render.getInstance();
		commands = Commands.getInstance();
	}

	@Override
	public void run() {
		while (alive) {
			timer = System.nanoTime();
			doLoop();
			render.update();
			timer += 16000000; // 62 fps
			// doit on quiter ?
			if (commands.keys.contains(Commands.QUIT))
				alive = false;
			while (System.nanoTime() < timer)
				Thread.yield();
		}
		System.exit(0);
	}

	private void doLoop() {
		frame++;
		// la boucle de jeu: on va gérer les etats du jeux ici et appeler les
		// fonctionnement adapté.
		switch (status) {
		case Const.INIT:
			frame = 0;
			invasion = 0;
			cash = Const.INI_CASH;
			waveDelai = Const.INI_WAVE_DELAY;
			mur.reset();
			render.resetLayers();
			render.getLayer(Render.TopLAYER).addRenderable(intro);
			status = Const.INTRO;
			break;
		case Const.INTRO:
			if (commands.isMousePressed() && frame > 10) {
				frame = 0;
				render.resetLayers();
				render.getLayer(Render.TopLAYER).addRenderable(briefing);
				status = Const.BRIEFING;
			}
			break;
		case Const.BRIEFING:
			if (commands.isMousePressed() && frame > 10) {
				ui.update(cash, invasion);
				render.resetLayers();
				render.getLayer(Render.MidLAYER).addRenderable(mur);
				render.getLayer(Render.UILAYER).addRenderable(ui);
				// a laisser apres l'ui ...
				render.getLayer(Render.UILAYER).addRenderable(select);
				status = Const.PLAY;
			}
			break;
		case Const.PLAY:
			doPlay();
			break;
		case Const.ENDGAME:
			if (commands.isMousePressed() && frame > 60) {
				status = Const.INIT;
			}
			break;
		default:
			break;
		}
	}
	/// --------------------------------------------------------------------------------------------
	// un jour ca serait pas mal de faire une classe qui gere ca ... pour le moment
	/// on avance

	private List<Somb> bots = new ArrayList<>();
	private List<ITower> towers = new ArrayList<>();
	private List<Somb> deadbots = null;
	private int actionIdx = -1;

	private void doPlay() {

		// toutes les secondes (60fps) on reduit le temps d'apparition des monstres (max
		// de 15 par seconde)
		if (frame % 60 == 0 && waveDelai > 4)
			waveDelai--;

		// On update la liste d'enemi
		if (frame % waveDelai == 0) {
			Somb monster = new Somb();
			bots.add(monster);
			render.getLayer(Render.BotLAYER).addRenderable(monster);
		}

		// toute les X ms Interaction tower / monstre
		if (frame % Const.SHOOT_DELAY == 0)
			for (ITower t : towers) {
				int round = t.round();
				for (Somb monster : bots) {
					if (t.isInRange(monster.x, monster.y)) {
						monster.hit(t.power());
						round--;
					}
					if (round < 0)
						break;
				}

			}

		// toute les 500ms on clean la liste de monstre
		if (frame % 30 == 0) {
			boolean cleaningWork = false;
			for (Somb monster : bots) {
				// on ne passe pas le mur (limite +epaisseur du mur)
				if (monster.x < Const.AREA_LEFT + 30 && monster.y < mur.getsize()) {
					monster.kill();
				}
				if (!monster.isActive()) {
					if (!cleaningWork) {
						cleaningWork = true;
						deadbots = new ArrayList<>();
					}
					deadbots.add(monster);
					render.getLayer(Render.BotLAYER).removeRenderable(monster);
					// le monstre est passé
					if (monster.isAlive()) {
						invasion += Const.INVASION_COST;
						if (invasion > Const.MAX_INVA)
							endgame(false);
					} else {
						// on l'a tué
						cash += monster.getValue();
					}
				}
			}
			if (cleaningWork) {
				bots.removeAll(deadbots);
				ui.update(cash, invasion);
			}
		}

		// on laisse un cycle pour l'affichage de la selection a la creation du mur
		if (actionIdx == Const.BT_MUR_IDX && frame % 10 == 0) {
			actionIdx = -1;
			select.reset();
			if (mur.getCost() > cash)
				ui.nocashBlink();
			else {
				cash -= mur.getCost();
				ui.update(cash, invasion);
				mur.add();
			}

			if (mur.isCompleted())
				endgame(true);
		}

		if (commands.isMouseClicked()) {
			// une action est demandé (si on est trop lent, on peut memoriser le clic
			// localement)

			// clic pour selectionner uner commande ?
			if (commands.cmx > Const.BT_MARGIN && commands.cmx < Const.BT_MARGIN + Const.BT_SIZE
					&& commands.cmy > Const.BT_OFFSET) {
				int y = commands.cmy - Const.BT_OFFSET;
				if (y % Const.BT_GAP < Const.BT_SIZE && y / Const.BT_GAP < Const.BT_LAST_IDX) {
					// le clic est valide et dans les limites du menu
					actionIdx = y / Const.BT_GAP;
					select.setPosition(actionIdx);
				}
			}
			// clic sur la carte en mode arosage!
			else if (actionIdx == Const.BT_RELATION_IDX && commands.cmx < 320 && commands.cmy < 200) {

				if (Const.PRICE_RELATION > cash)
					ui.nocashBlink();
				else {
					cash -= Const.PRICE_RELATION;
					ui.update(cash, invasion);
					ui.calm();
					// On pourrai ne pas debiter le compte en banque si tout est calme ... ou pas
				}
			}
			// le joueur cherche a poser une tour, la zone est valide
			else if (actionIdx > 1 && commands.cmx > Const.AREA_LEFT && commands.cmy > Const.AREA_TOP) {
				ITower tower = null;
				switch (actionIdx) {
				case Const.BT_BARB_IDX:
					tower = new TowerBarbed(commands.cmx, commands.cmy);
					break;
				case Const.BT_TOWER_IDX:
					tower = new TowerWatch(commands.cmx, commands.cmy);
					break;
				case Const.BT_HUMMER_IDX:
					tower = new TowerHummer(commands.cmx, commands.cmy);
					break;
				}
				if (tower != null) {
					if (tower.getCost() > cash)
						ui.nocashBlink();
					else {
						cash -= tower.getCost();
						ui.update(cash, invasion);
						towers.add(tower);
						render.getLayer(Render.BotLAYER).addRenderable(tower);
						ui.panic(tower.getCost());
						// si c'est la pannique general on a perdu
						if (ui.checkmap())
							endgame(false);
					}
				}

			} else {
				actionIdx = -1;
				select.reset();
			}

		}

	}

	// traitement victoire et defaite
	private void endgame(boolean victory) {
		this.status = Const.ENDGAME;
		frame = 0;
		render.resetLayers();
		render.getLayer(Render.MidLAYER).addRenderable(victory ? win : loose);
	}
}
